"""
This submodule provides several software/hardware datasets.

These examples provide quick access to simulated control signal sequences from
hardware prototypes. Additionally, as simulating a control-bounded ADC is
typically a computationally intense task, compared to filtering, several control
signal sequences from default architectures and parametrizations are provided to
reduce simulation time.
"""

# from .hadamard import HadamardPCB
# from .chain_of_integrators import ChainOfIntegrators
